@@ -28,8 +28,9 @@
 #include <atomic>
 #include <mutex>
 #include "WFTaskError.h"
-#include "WFKafkaClient.h"
 #include "StringUtil.h"
+#include "KafkaTaskImpl.inl"
+#include "WFKafkaClient.h"
 
 #define KAFKA_HEARTBEAT_INTERVAL	(3 * 1000 * 1000)
 
@@ -313,7 +314,10 @@ void KafkaClientTask::kafka_rebalance_callback(__WFKafkaTask *task)
 		WFTaskFactory::signal_by_name(name, NULL, max);
 	}
 	else
+	{
 		kafka_rebalance_proc(member, series);
+		member->mutex.unlock();
+	}
 }
 
 void KafkaClientTask::kafka_rebalance_proc(KafkaMember *member, SeriesWork *series)
@@ -336,8 +340,6 @@ void KafkaClientTask::kafka_rebalance_proc(KafkaMember *member, SeriesWork *seri
 	member->cgroup_outdated = false;
 
 	series->push_back(task);
-
-	member->mutex.unlock();
 }
 
 void KafkaClientTask::kafka_heartbeat_callback(__WFKafkaTask *task)
@@ -355,12 +357,7 @@ void KafkaClientTask::kafka_heartbeat_callback(__WFKafkaTask *task)
 		return;
 	}
 
-	if (resp->get_cgroup()->get_error() != 0)
-	{
-		kafka_rebalance_proc(member, series);
-		return;
-	}
-	else
+	if (resp->get_cgroup()->get_error() == 0)
 	{
 		member->heartbeat_status = KAFKA_HEARTBEAT_DONE;
 		WFTimerTask *timer_task;
@@ -369,6 +366,8 @@ void KafkaClientTask::kafka_heartbeat_callback(__WFKafkaTask *task)
 		timer_task->user_data = member;
 		series->push_back(timer_task);
 	}
+	else
+		kafka_rebalance_proc(member, series);
 
 	member->mutex.unlock();
 }
@@ -1209,6 +1208,9 @@ bool KafkaClientTask::add_toppar(const KafkaToppar& toppar)
 		}
 
 		new_toppar.set_offset(toppar.get_offset());
+		new_toppar.set_offset_timestamp(toppar.get_offset_timestamp());
+		new_toppar.set_low_watermark(toppar.get_low_watermark());
+		new_toppar.set_high_watermark(toppar.get_high_watermark());
 		this->toppar_list.add_item(new_toppar);
 
 		this->meta_list.add_item(tmp);
@@ -1239,6 +1241,9 @@ bool KafkaClientTask::add_toppar(const KafkaToppar& toppar)
 		}
 
 		new_toppar.set_offset(toppar.get_offset());
+		new_toppar.set_offset_timestamp(toppar.get_offset_timestamp());
+		new_toppar.set_low_watermark(toppar.get_low_watermark());
+		new_toppar.set_high_watermark(toppar.get_high_watermark());
 		this->toppar_list.add_item(new_toppar);
 
 		this->meta_list.add_item(*meta);
@@ -1450,7 +1455,9 @@ int KafkaClientTask::arrange_fetch()
 						return -1;
 
 					new_toppar.set_offset(toppar->get_offset());
+					new_toppar.set_offset_timestamp(toppar->get_offset_timestamp());
 					new_toppar.set_low_watermark(toppar->get_low_watermark());
+					new_toppar.set_high_watermark(toppar->get_high_watermark());
 					this->toppar_list_map[node_id].add_item(std::move(new_toppar));
 				}
 			}
